@@ -19,17 +19,22 @@ logging.basicConfig(
 )
 
 import time
+import types
+import typing
 from typing import (
     Any,
     Dict,
     List,
+    Union,
     TypeVar,
     Callable,
     Iterable,
+    Optional,
+    Protocol,
     Generator,
-    TypeAlias,
     AsyncGenerator,
 )
+from inspect import signature
 from concurrent import futures
 from dataclasses import field, dataclass
 
@@ -41,16 +46,17 @@ from google.protobuf import json_format
 from service_pb2_grpc import OrcaProcessorServicer
 
 from orca_python import envs
-from orca_python.exceptions import InvalidDependency, InvalidAlgorithmArgument
+from orca_python.exceptions import (
+    InvalidDependency,
+    InvalidAlgorithmArgument,
+    InvalidAlgorithmReturnType,
+)
 
 # Regex patterns for validation
 ALGORITHM_NAME = r"^[A-Z][a-zA-Z0-9]*$"
 SEMVER_PATTERN = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)$"
 WINDOW_NAME = r"^[A-Z][a-zA-Z0-9]*$"
 
-AlgorithmFn: TypeAlias = Callable[..., Any]
-
-T = TypeVar("T", bound=AlgorithmFn)
 
 LOGGER = logging.getLogger(__name__)
 
@@ -74,6 +80,74 @@ class WindowType:
             )
 
 
+@dataclass
+class StructResult:
+    value: Dict[str, Any]
+
+    def __init__(self, value: Dict[str, Any]) -> None:
+        """
+        Produce a struct/dictionary based result
+
+        Args:
+            value: The result to produce. e.g.: {'min': -1.1, 'median': 4.2, 'max': 5.0}
+        """
+        self.value = value
+
+
+@dataclass
+class ValueResult:
+    value: float | int | bool
+
+    def __init__(self, value: float | int | bool) -> None:
+        """
+        Produce a value result
+
+        Args:
+            value: The result to produce. E.g. 1.0
+        """
+        self.value = value
+
+
+@dataclass
+class ArrayResult:
+    value: Iterable[float | int | bool]
+
+    def __init__(self, value: Iterable[float | int | bool]) -> None:
+        """
+        Produce an array result
+
+        Args:
+            value: The result to produce. E.g. [1, 2, 3, 4, 5]
+        """
+        self.value = value
+
+
+class NoneResult:
+    """
+    The `None` result type
+    """
+
+    value: None = None
+
+
+returnResult = StructResult | ArrayResult | ValueResult | NoneResult
+
+
+@dataclass
+class ExecutionParams:
+    window: pb.Window
+    dependencies: Optional[Iterable[pb.AlgorithmResult]] = None
+
+
+class AlgorithmFn(Protocol):
+    def __call__(
+        self, params: ExecutionParams, *args: Any, **kwargs: Any
+    ) -> returnResult: ...
+
+
+T = TypeVar("T", bound=AlgorithmFn)
+
+
 @dataclass
 class Window:
     time_from: int
@@ -134,6 +208,7 @@ class Algorithm:
     exec_fn: AlgorithmFn
     processor: str
     runtime: str
+    result_type: returnResult
 
     @property
     def full_name(self) -> str:
@@ -200,12 +275,11 @@ class Algorithms:
                 break
 
         if not dependencyAlgo:
+            dep_name = getattr(dependency, "__name__", "<unknown>")
             LOGGER.error(
-                f"Failed to find registered algorithm for dependency: {dependency.__name__}"
-            )
-            raise ValueError(
-                f"Dependency {dependency.__name__} not found in registered algorithms"
+                f"Failed to find registered algorithm for dependency: {dep_name}"
             )
+            raise ValueError(f"Dependency {dep_name} not found in reg=ms")
 
         if algorithm not in self._dependencyFns:
             self._dependencyFns[algorithm] = [dependency]
@@ -265,7 +339,7 @@ class Processor(OrcaProcessorServicer):  # type: ignore
         self,
         exec_id: str,
         algorithm: pb.Algorithm,
-        dependencyResults: Iterable[pb.AlgorithmResult],
+        params: ExecutionParams,
     ) -> pb.ExecutionResult:
         """
         Executes a single algorithm with resolved dependencies.
@@ -273,7 +347,7 @@ class Processor(OrcaProcessorServicer):  # type: ignore
         Args:
             exec_id (str): Unique execution ID.
             algorithm (pb.Algorithm): The algorithm to execute.
-            dependencyResults (Iterable[pb.AlgorithmResult]): Results from dependency algorithms.
+            params (ExecutionParams): The execution params object, which contains the triggering window and dependency results.
 
         Returns:
             pb.ExecutionResult: The result of the execution.
@@ -288,27 +362,28 @@ class Processor(OrcaProcessorServicer):  # type: ignore
 
             # convert dependency results into a dict of name -> value
             dependency_values = {}
-            for dep_result in dependencyResults:
-                # extract value based on which oneof field is set
-                dep_value = None
-                if dep_result.result.HasField("single_value"):
-                    dep_value = dep_result.result.single_value
-                elif dep_result.result.HasField("float_values"):
-                    dep_value = list(dep_result.result.float_values.values)
-                elif dep_result.result.HasField("struct_value"):
-                    dep_value = json_format.MessageToDict(
-                        dep_result.result.struct_value
-                    )
+            if params.dependencies:
+                for dep_result in params.dependencies:
+                    # extract value based on which oneof field is set
+                    dep_value = None
+                    if dep_result.result.HasField("single_value"):
+                        dep_value = dep_result.result.single_value
+                    elif dep_result.result.HasField("float_values"):
+                        dep_value = list(dep_result.result.float_values.values)
+                    elif dep_result.result.HasField("struct_value"):
+                        dep_value = json_format.MessageToDict(
+                            dep_result.result.struct_value
+                        )
 
-                dep_name = f"{dep_result.algorithm.name}_{dep_result.algorithm.version}"
-                dependency_values[dep_name] = dep_value
+                    dep_name = (
+                        f"{dep_result.algorithm.name}_{dep_result.algorithm.version}"
+                    )
+                    dependency_values[dep_name] = dep_value
 
             # execute in thread pool since algo.exec_fn is synchronous
             loop = asyncio.get_event_loop()
 
-            algoResult = await loop.run_in_executor(
-                None, algo.exec_fn, dependency_values
-            )
+            algoResult = await loop.run_in_executor(None, algo.exec_fn, params)
 
             # create result based on the return type
             current_time = int(time.time())  # Current timestamp in seconds
@@ -410,13 +485,13 @@ class Processor(OrcaProcessorServicer):  # type: ignore
             return pb.ExecutionResult(exec_id=exec_id, algorithm_result=algo_result)
 
     def ExecuteDagPart(
-        self, ExecutionRequest: pb.ExecutionRequest, context: grpc.ServicerContext
+        self, executionRequest: pb.ExecutionRequest, context: grpc.ServicerContext
     ) -> Generator[pb.ExecutionResult, None, None]:
         """
         Executes part of a DAG (Directed Acyclic Graph) of algorithms.
 
         Args:
-            ExecutionRequest (pb.ExecutionRequest): The DAG execution request.
+            executionRequest (pb.ExecutionRequest): The DAG execution request.
             context (grpc.ServicerContext): gRPC context for the request.
 
         Yields:
@@ -428,8 +503,8 @@ class Processor(OrcaProcessorServicer):  # type: ignore
 
         LOGGER.info(
             (
-                f"Received DAG execution request with {len(ExecutionRequest.algorithms)} "
-                f"algorithms and ExecId: {ExecutionRequest.exec_id}"
+                f"Received DAG execution request with {len(executionRequest.algorithms)} "
+                f"algorithms and ExecId: {executionRequest.exec_id}"
             )
         )
 
@@ -444,11 +519,14 @@ class Processor(OrcaProcessorServicer):  # type: ignore
             # create tasks for all algorithms
             tasks = [
                 self.execute_algorithm(
-                    ExecutionRequest.exec_id,
+                    executionRequest.exec_id,
                     algorithm,
-                    ExecutionRequest.algorithm_results,
+                    ExecutionParams(
+                        window=executionRequest.window,
+                        dependencies=executionRequest.algorithm_results,
+                    ),
                 )
-                for algorithm in ExecutionRequest.algorithms
+                for algorithm in executionRequest.algorithms
             ]
 
             # execute all tasks concurrently and yield results as they complete
@@ -526,6 +604,21 @@ class Processor(OrcaProcessorServicer):  # type: ignore
             algo_msg.name = algorithm.name
             algo_msg.version = algorithm.version
 
+            # manage the return type of the algorithm
+            if algorithm.result_type == ValueResult:  # type: ignore
+                result_type_pb = pb.ResultType.VALUE
+            elif algorithm.result_type == StructResult:  # type: ignore
+                result_type_pb = pb.ResultType.STRUCT
+            elif algorithm.result_type == ArrayResult:  # type: ignore
+                result_type_pb = pb.ResultType.ARRAY
+            else:
+                raise InvalidAlgorithmReturnType(
+                    f"Algorithm has return type {algorithm.result_type}, but expected one of `StructResult`, `ValueResult`, `ArrayResult`"
+                )
+
+            ## add the result type
+            algo_msg.result_type = result_type_pb
+
             # Add window type
             algo_msg.window_type.name = algorithm.window_name
             algo_msg.window_type.version = algorithm.window_version
@@ -637,18 +730,16 @@ class Processor(OrcaProcessorServicer):  # type: ignore
 
         def inner(algo: T) -> T:
             def wrapper(
-                dependency_values: Dict[str, Any] | None = None,
+                params: ExecutionParams,
                 *args: Any,
                 **kwargs: Any,
-            ) -> Any:
+            ) -> returnResult:
                 LOGGER.debug(f"Executing algorithm {name}_{version}")
                 try:
                     # setup ready for the algo
-                    # add dependency values to kwargs if provided
-                    if dependency_values:
-                        kwargs["dependencies"] = dependency_values
+                    # pack the params into the kwargs - user can decide if they want it
+                    kwargs["params"] = params
                     LOGGER.debug(f"Algorithm {name}_{version} setup complete")
-                    # TODO
 
                     # run the algo
                     LOGGER.info(f"Running algorithm {name}_{version}")
@@ -664,6 +755,13 @@ class Processor(OrcaProcessorServicer):  # type: ignore
                     )
                     raise
 
+            sig = signature(algo)
+            returnType = sig.return_annotation
+            if not is_type_in_union(returnType, returnResult):  # type: ignore
+                raise InvalidAlgorithmReturnType(
+                    f"Algorithm has return type {sig.return_annotation}, but expected one of `StructResult`, `ValueResult`, `ArrayResult`"
+                )
+
             algorithm = Algorithm(
                 name=name,
                 version=version,
@@ -673,6 +771,7 @@ class Processor(OrcaProcessorServicer):  # type: ignore
                 exec_fn=wrapper,
                 processor=self._name,
                 runtime=sys.version,
+                result_type=returnType,
             )
 
             self._algorithmsSingleton._add_algorithm(algorithm.full_name, algorithm)
@@ -695,6 +794,29 @@ class Processor(OrcaProcessorServicer):  # type: ignore
             # needs to be defined before a dependency can be created, and you can only register depencenies
             # once. But when dependencies are grabbed from a server, circular dependencies will be possible
 
-            return wrapper  # type: ignore
+            return wrapper  # type: ignore[return-value]
 
         return inner
+
+
+def is_type_in_union(target_type, union_type):  # type: ignore
+    """
+    Check if target_type is contained within union_type.
+    Works with both new syntax (int | float) and typing.Union.
+    """
+    try:
+        # handle new union syntax (Python 3.10+) - types.UnionType
+        if isinstance(union_type, types.UnionType):
+            return target_type in union_type.__args__
+
+        # handle typing.Union syntax
+        origin = getattr(typing, "get_origin", lambda x: None)(union_type)
+        if origin is Union:
+            args = getattr(typing, "get_args", lambda x: ())(union_type)
+            return target_type in args
+
+        # handle single type (not a union)
+        return target_type == union_type
+
+    except (AttributeError, TypeError):
+        return False
